/* File listmanip.sig. Various simple list manipulation programs
needed for writing our theorem provers are defined here.  All of these
programs are essentially first-order and correspond to code one would
write in normal Prolog.  */

module listmanip.

type    member                  A -> (list A) -> o.
type    member_and_rest         A -> (list A) -> (list A) -> o.
type    nth_item                int -> A -> (list A) -> o.
type    nth_item_and_rest       int -> A -> (list A) -> (list A) -> o.
type    member_move_to_end      A -> (list A) -> (list A) -> o.
type    add_to_end              A -> (list A) -> (list A) -> o.
type    list_diff               (list A) -> (list A) -> (list A) -> o.
type    append  		(list A) -> (list A) -> (list A) -> o.
type    nth_item_replace        int -> A -> (list A) -> (list A)
	     		         -> (list A) -> o.
type    length		        (list A) -> int -> o.
type    divide_list	        (list int) -> (list A)
				      -> (list A) -> (list A) -> o.
type    nth_item_divide_rest    int -> (list int) -> A -> (list A)
			         -> (list A) -> (list A) -> o.

local   divide_list_aux         int -> int -> (list int) -> (list A)
			         -> (list A) -> (list A) -> o.
local   nth_item_divide_aux     int -> int -> int -> (list int) -> A -> (list A)
			         -> (list A) -> (list A) -> o.


member X (X::L) :- !.
member X (Y::L) :- member X L.

member_and_rest A (A::Rest) Rest.
member_and_rest A (B::Tail) (B::Rest) :-
  member_and_rest A Tail Rest.

nth_item 0 A List :- !, member A List.
nth_item 1 A (A::Rest) :- !.
nth_item N A (B::Tail) :-
  (N1 is (N - 1)), nth_item N1 A Tail.

nth_item_and_rest 0 A List Rest :- !,
  member_and_rest A List Rest.
nth_item_and_rest 1 A (A::Rest) Rest.
nth_item_and_rest N A (B::Tail) (B::Rest) :-
  (N1 is (N - 1)),
  nth_item_and_rest N1 A Tail Rest.

member_move_to_end A (A::Rest) NewList :-
  add_to_end A Rest NewList.
member_move_to_end A (B::Tail) (B::NewList) :-
  member_move_to_end A Tail NewList.

add_to_end A nil (A::nil).
add_to_end A (Head::Tail) (Head::NewTail) :-
  add_to_end A Tail NewTail.

list_diff nil _ nil :- !.
list_diff (A::Rest) List NewList :-
  member A List, !, list_diff Rest List NewList.
list_diff (A::Rest) List (A::NewList) :-
  list_diff Rest List NewList.

append nil L L :- !.
append (A::Rest) L (A::NewList) :-
  append Rest L NewList.

nth_item_replace 1 A L (A::Rest) NewList :- !,
  append L Rest NewList.
nth_item_replace N A L (C::Rest) (C::NewList) :-
  (N1 is (N - 1)),
  nth_item_replace N1 A L Rest NewList.

length nil 0.
length (A::Rest) N :- length Rest N1, (N is (N1 + 1)).

divide_list Ns L NewList1 NewList2 :-
  length L Nl,
  divide_list_aux 1 Nl Ns L NewList1 NewList2.

divide_list_aux Start End Ns L nil nil :-
  Start > End, !.
divide_list_aux Start End Ns (B::Rest) (B::Rest1) Rest2 :-
   member Start Ns, !,
  (Start1 is (Start + 1)), 
  divide_list_aux Start1 End Ns Rest Rest1 Rest2.
divide_list_aux Start End Ns (B::Rest) Rest1 (B::Rest2) :-
  (Start1 is (Start + 1)), 
  divide_list_aux Start1 End Ns Rest Rest1 Rest2.

nth_item_divide_rest N Ns A L Rest1 Rest2 :-
  length L Nl,
  nth_item_divide_aux 1 N Nl Ns A L Rest1 Rest2.

nth_item_divide_aux Start N End Ns A L nil nil :-
  Start > End, !.
nth_item_divide_aux Start N End Ns A (A::Rest) Rest1 Rest2 :-
  N = Start, !,
  (Start1 is (Start + 1)), 
  nth_item_divide_aux Start1 N End Ns A Rest Rest1 Rest2.
nth_item_divide_aux Start N End Ns A (B::Rest) (B::Rest1) Rest2 :-
   member Start Ns, !,
  (Start1 is (Start + 1)), 
  nth_item_divide_aux Start1 N End Ns A Rest Rest1 Rest2.
nth_item_divide_aux Start N End Ns A (B::Rest) Rest1 (B::Rest2) :-
  (Start1 is (Start + 1)), 
  nth_item_divide_aux Start1 N End Ns A Rest Rest1 Rest2.


